(*

LockFreePrimitives

Description:
 Ring Buffer Test

Homepage:
 http://www.idescriptor.com/RingBuffer/RingBuffer.htm

Mail:
 gorazd.jernejc@gmail.com

Copyright (c) 2009, Gorazd Jernejc
All rights reserved.

License:
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
 * The name of the Gorazd Jernejc may not be used to endorse or promote products
   derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 COMPANIES WHOSE NAME STARTS WITH LETTERS 'DANF' MAY NOT USE THIS SOFTWARE!

History:
 Author            : GJ
 Creation date     : 31. August 2009
 Last modification : 31. August 2009
 Version           : 1.00

 1.00: 31. August 2009
 First official release.
*)

unit rbTest;

{$R-,O+,A8}

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Forms, Dialogs, StdCtrls, Spin,
  RingBuffer,
  LockFreePrimitives;

const
 MSG_WRITER_COMPLETED = WM_USER + 1;
 MSG_READER_COMPLETED = WM_USER + 2;

type
  TWriteThread = class(TThread)
  protected
    Index: integer;
    procedure Execute; override;
  public
  end;

  TReadThread = class(TThread)
  protected
    Index: integer;
    procedure Execute; override;
  public
  end;

  PMsgRecord = ^TMsgRecord;
  TMsgRecord = record
    Count           : int64;
    EndTickCount    : int64;
    Index           : integer;
    IOCounter       : int64;
    StartTickCount  : int64;
    rbFullEmptyCount: int64;
  end;

  TrbForm = class(TForm)
    btnStartTest: TButton;
    seWritersCount: TSpinEdit;
    LabelWriters: TLabel;
    ListLog: TListBox;
    seLinksPerWriter: TSpinEdit;
    Label1: TLabel;
    seRingBufferLength: TSpinEdit;
    Labelrbl: TLabel;
    seReadersCount: TSpinEdit;
    Label2: TLabel;
    btnClear: TButton;
    procedure btnClearClick(Sender: TObject);
    procedure btnStartTestClick(Sender: TObject);
  private
    { Private declarations }
    AllReadersCount          : int64;
    AllTasksFinishedTick     : int64;
    AllWritersCount          : int64;
    ReaderTaskFinishedCounter: integer;
    ReadThread               : array[1..128]of TReadThread;
    TasksStarted             : int64;
    WriterTaskFinishedCounter: integer;
    WriteThread              : array[1..128]of TWriteThread;
    procedure Log(const msg: string);
    procedure ReaderFinished(var Message: TMessage); message MSG_READER_COMPLETED;
    procedure WriterFinished(var Message: TMessage); message MSG_WRITER_COMPLETED ;
  public
    { Public declarations }
  end;

var
  rbForm        : TrbForm;
  rb            : TgjRingBuffer;
  qpf           : int64;
  TestCounter   : cardinal = 0;
  //Global variables
  LinksPerWriter: integer;
  ReadersCount  : integer;
  TaskSuspended : integer;
  StartCounter  : integer;
  StartTest     : boolean;
  WritersCount  : integer;


implementation

{$R *.dfm}

function spExt(const num: integer): string;
begin
  if num > 1 then
    result := 's' else
    result := '';
end;

procedure TrbForm.btnClearClick(Sender: TObject);
begin
  ListLog.Clear;
  TestCounter := 0;
end;

procedure TrbForm.btnStartTestClick(Sender: TObject);
var
  n: integer;
begin
  WritersCount := seWritersCount.Value;
  ReadersCount := seReadersCount.Value;
  LinksPerWriter := seLinksPerWriter.Value;
  AllReadersCount := 0;
  AllWritersCount := 0;
  ReaderTaskFinishedCounter := 0;
  WriterTaskFinishedCounter := 0;
  StartCounter := 0;
  StartTest := False;
  //Test start
  Log('');
  inc(TestCounter);
  Log('Starting test number: ' + IntTostr(TestCounter));
  btnStartTest.Enabled := False;
  Log('Creating ring buffer ' +
    FloatToStrF(seRingBufferLength.Value, ffNumber, 10, 0) + ' links depth...');
  TaskSuspended := 0;
  rb := TgjRingBuffer.Create;
  rb.Initialize(seRingBufferLength.Value);
  Log('Creating ' + IntToStr(WritersCount) + ' writer'+ spExt(WritersCount) + '...');
  for n := 1 to WritersCount do
  begin
    WriteThread[n] := TWriteThread.Create(True);
    WriteThread[n].Index := n;
    WriteThread[n].Resume;
  end;
  Log('Creating ' + IntToStr(ReadersCount) +  ' reader'+ spExt(ReadersCount) + '...');
  for n := 1 to ReadersCount do
  begin
    ReadThread[n] := TReadThread.Create(True);
    ReadThread[n].Index := n;
    ReadThread[n].Resume;
  end;
  while not StartCounter = WritersCount + ReadersCount do
    SwitchToThread;
  Log('Starting test with ' + FloatToStrF(LinksPerWriter, ffNumber, 10, 0) + ' links per writer...');
  //Relax OS kernel before start test
  Sleep(100);
  //Start test
  QueryPerformanceCounter(TasksStarted);
  StartTest := True;
  SFence;
end;

procedure TrbForm.Log(const msg: string);
begin
  ListLog.ItemIndex := ListLog.Items.Add(msg);
end;

procedure TrbForm.ReaderFinished(var Message: TMessage); var n: integer;
begin
  with  PMsgRecord(Message.LParam)^ do begin
    if Message.WParam <> Index then
      raise Exception.Create('Error: Therad and Message Index not EQU.');
    Log('Reader number ' + IntToStr(Index) + ' red ' + FloatToStrF(IOCounter, ffNumber, 10, 0) +
      ' msg; Bandwith ' +  FloatToStrF((IOCounter / (EndTickCount - StartTickCount) * qpf),
        ffNumber, 10, 0) + ' msg/s;' +
      ' Buffer empty count: ' + FloatToStrF(rbFullEmptyCount, ffNumber, 10, 0) + ';');
    inc(AllReadersCount, IOCounter);
    inc(ReaderTaskFinishedCounter);
    FreeMem(PMsgRecord(Message.LParam));
    if ReaderTaskFinishedCounter >= ReadersCount then begin
      QueryPerformanceCounter(AllTasksFinishedTick);
      Log('All readers red ' + FloatToStrF(AllReadersCount, ffNumber, 10, 0) + ' msg;');
      Log('Common bandwith: ' + FloatToStrF(AllReadersCount /
        (AllTasksFinishedTick - TasksStarted) * qpf, ffNumber, 10, 0) + ' msg/s;');
      for n := 1 to WritersCount do
        FreeAndNil(WriteThread[n]);
      for n := 1 to ReadersCount do
        FreeAndNil(ReadThread[n]);
      rb.Free;
      Log('Test stoped!');
      btnStartTest.Enabled := True;
    end;
  end;
end;

procedure TrbForm.WriterFinished(var Message: TMessage);
begin
  with  PMsgRecord(Message.LParam)^ do begin
    Log('Writer number ' + IntToStr(Index) + ' sent ' + FloatToStrF(IOCounter, ffNumber, 10, 0) +
      ' msg; Bandwith ' +  FloatToStrF((IOCounter / (EndTickCount - StartTickCount) * qpf),
        ffNumber, 10, 0) + ' msg/s;' +
      ' Buffer full count: ' + FloatToStrF(rbFullEmptyCount, ffNumber, 10, 0) + ';');
    inc(AllWritersCount, IOCounter);
    inc(WriterTaskFinishedCounter);
    FreeMem(PMsgRecord(Message.LParam));
    if WriterTaskFinishedCounter >= WritersCount then begin
      Log('All writers sent ' + FloatToStrF(AllWritersCount, ffNumber, 10, 0) + ' msg;');
    end;
  end;
end;

{ TWriteThread }

procedure TWriteThread.Execute;
var
  Msg: PMsgRecord;
  n  : cardinal;
begin
  //Create Msg record
  GetMem(Msg, SizeOf(TMsgRecord));
  with Msg^ do begin
    Index := self.Index;
    rbFullEmptyCount := 0;
    //Signale to main process that thread is ready to start
    InterlockedIncrement(StartCounter);
    //Wait on global start signal StartTest
    while not StartTest do
      SwitchToThread;
    //Thread task...
    IOCounter := 0;
    QueryPerformanceCounter(StartTickCount);
    for n := 1 to LinksPerWriter do
    begin
      while rb.InsertRight(pointer(n)) = 0 do begin
       inc(rbFullEmptyCount);
       SwitchToThread;
      end;
      inc(IOCounter);
    end;
    //Thread task finished...
    QueryPerformanceCounter(EndTickCount);
    InterlockedIncrement(TaskSuspended);
    PostMessage(rbForm.Handle, MSG_WRITER_COMPLETED, self.Index , Integer(msg));
    //Do not overcharge the OS kernel so...
    Suspend;
  end;
end;

{ TReadThread }

procedure TReadThread.Execute;
var
  Msg: PMsgRecord;
  n  : cardinal;
begin
  //Create Msg record
  GetMem(Msg, SizeOf(TMsgRecord));
  with Msg^ do begin
    Index := self.Index;
    rbFullEmptyCount := 0;
    //Signale to main process that thread is ready to start
    InterlockedIncrement(StartCounter);
    //Wait on global start signal StartTest
    while not StartTest do
      SwitchToThread;
    //Thread task...
    IOCounter := 0;
    QueryPerformanceCounter(StartTickCount);
    repeat
      if rb.RemoveLeft(pointer(n)) <> 0 then
        inc(IOCounter) else
        if (TaskSuspended = WritersCount) and (rb.Count = 0) then
          break else begin
            inc(rbFullEmptyCount);
            SwitchToThread;
          end;
    until false;
    //Thread task finished...
    QueryPerformanceCounter(EndTickCount);
    PostMessage(rbForm.Handle, MSG_READER_COMPLETED, self.Index, Integer(msg));
    //Do not overcharge OS kernel so...
    Suspend;
  end;
end;

initialization
  QueryPerformanceFrequency(qpf);

end.
